@@ -1411,7 +1411,6 @@ static int test_kernel_modules_helper(Options *op, Package *p, int pause_udev)
         "drm",
         "drm-kms-helper",
         "drm-ttm-helper",
-        "drm_client_lib",
         "vfio_mdev",
         "vfio",
         "mdev",
@@ -1420,6 +1419,11 @@ static int test_kernel_modules_helper(Options *op, Package *p, int pause_udev)
         "vfio_pci_core",
         "ecc",
         "nvgrace-egm",
+        // Modules for Tegra
+        "tsecriscv",
+        "tegra_dce",
+        "host1x_nvhost",
+        "mc_utils",
     };
 
     if (pause_udev) {
